{% extends 'generic/object_list.html' %}
{% load buttons %}
{% load plugins %}
{% load i18n %}

{% block controls %}
  <div class="btn-list">
    {% plugin_list_buttons model %}
    {% action_buttons actions model %}
    <a href="#"
       hx-get="{% url 'plugins:ipfabric_netbox:ipfabrictransformmap_restore' %}"
       hx-target="#htmx-modal-content"
       class="btn btn-warning"
       data-bs-toggle="modal"
       data-bs-target="#htmx-modal">
      <span class="mdi mdi-file-restore" aria-hidden="true"></span>&nbsp;{% trans "Restore" %}
    </a>
    <a href="{% url 'plugins:ipfabric_netbox:ipfabrictransformmap_import_json' %}"
       class="btn btn-cyan">
      <span class="mdi mdi-upload" aria-hidden="true"></span>&nbsp;{% trans "Import JSON" %}
    </a>
    <a href="{% url 'plugins:ipfabric_netbox:ipfabrictransformmap_export_json' %}"
       class="btn btn-purple">
      <span class="mdi mdi-download" aria-hidden="true"></span>&nbsp;{% trans "Export JSON" %}
    </a>
  </div>
{% endblock controls %}

{% block modals %}
  {% include 'inc/htmx_modal.html' %}
{% endblock modals %}
