{% load i18n %}
{# Indicates that a job result is still pending; used for HTMX requests #}
<div class="spinner-border float-start me-1" id="spinner" role="status">
  <span class="visually-hidden">{% trans "Loading..." %}</span>
</div>
<h3>{% trans "Logs pending..." %}</h3>
<small class="text-muted">{% trans "Last updated" %} {% now "H:i:s" %}</small>
