# Generated by Django 5.2.11 on 2026-03-05 19:02
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0030_ipfabricsync_thread_count"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricingestionissue",
            name="phase",
            field=models.CharField(default="sync", max_length=10),
        ),
        migrations.AddField(
            model_name="ipfabricingestion",
            name="merge_job",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="merge_ingestion",
                to="core.job",
            ),
        ),
    ]
