# Generated by Django 5.2 on 2025-07-11 19:39
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        (
            "ipfabric_netbox",
            "0014_ipfabrictransformmapgroup_ipfabrictransformmap_group",
        ),
    ]

    operations = [
        migrations.CreateModel(
            name="IPFabricIngestionIssue",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("timestamp", models.DateTimeField(default=django.utils.timezone.now)),
                ("model", models.CharField(blank=True, max_length=100, null=True)),
                ("message", models.TextField()),
                ("raw_data", models.TextField(blank=True, default="")),
                ("coalesce_fields", models.TextField(blank=True, default="")),
                ("defaults", models.TextField(blank=True, default="")),
                ("exception", models.TextField()),
                (
                    "ingestion",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="issues",
                        to="ipfabric_netbox.ipfabricingestion",
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Ingestion Issue",
                "verbose_name_plural": "IP Fabric Ingestion Issues",
                "ordering": ["timestamp"],
            },
        ),
    ]
